#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdarg.h>
#include <time.h>
#include <limits.h>

#include "tapeio.h"
#include "tap.h"


#define TAPE_BLOCK_MAX_SIZE 65000U	// maximum bytes per block of tape
#define BLOCKING_FACTOR 1U		// number of blocks per pass
#define TAPE_BUFFER_SIZE TAPE_BLOCK_MAX_SIZE  // buffer size allocated
#define VERSION "0.02"			// version number
#define COMMENT_LINE_MAX 80		// length of longest comment line
#define COMMENT_SEPARATOR "\n---------------------------------------------\n"
#define LOG_FILE_TYPE ".LOG"
#define TAP_FILE_TYPE ".TAP"

//  Prototypes.

void FlushRecordCount( void);
void Logprintf( char *Format, ...);
void PrintUsage( void);

//  Globals.

char
  ProgramName[ _MAX_FNAME+1],
  TapFile[_MAX_PATH+1],
  LogFile[_MAX_PATH+1];

static const char *LogonText =
  "\n"
  "                     %s - Read SCSI Tape to TAP file\n\n"
  "                       Copyright 2015, Sydex, Inc.\n"
  "                         All rights reserved.\n\n";

static const char *UsageText =
  " %s - Read tape into SIMH .TAP file\n\n"
  " Command line syntax is:\n\n"
  "    %s [-r] [-c] [-n count] tape-file\n\n"
  " If -r is specified, suppresses all tape rewinding.\n"
  " If -c is specified, appends log file data after EOT marker.\n"
  " If -n count is specified, copies only \"count\" blocks.\n"
  " If -t is specified, stop at 2 consecutive tape marks.\n"
  " \"tape-file\" is the base file name; file types are suffixed.\n"
  " A file type of .TAP is used for the image file.\n"
  " A file type of .LOG is used for the logging file.\n\n";

BYTE
  _far *SourceBuffer;			// buffer for reading

WORD
  LastRecordLength,			// length of last record
  LastRecordCount,			// last record count
  ActualRead;				// how many bytes in record

DWORD
  BlocksToCopy,				// copy how many records?
  CurrentLength,			// current length for header
  BytesCopied;				// how many total bytes read

int
  SourceTape,				// what tape is being used
  TapeCount,				// how many tapes are there
  AppendLog,				// if nonzero, add comment after EOT
  NoRewind,				// if nonzero, don't ever rewind
  TwoMarks;				// if nonzero, stop at two tapemarks

FILE
    *Tap,			       // our output (TAP) file
    *Log;			       // temporary - holds trailing data

//  Tape Read Program.
//  ------------------
//
//  Locates tape drives, selects one and then reads and writes the
//  contents as a SIMH .TAP file.
//

void main( int argc, char *argv[])
{

  int
    status,
    tapemarkHit;

  time_t
    currTime;				// for time funcitons

  char
    commentLine[ COMMENT_LINE_MAX+1];	// commend buffer
  char
    *ch;				// temporary

//  Look at the command line and grab the command name.

  _splitpath( argv[0], NULL, NULL, ProgramName, NULL);
  strupr( ProgramName);
  argv++;
  argc--;

//  Digest any options.

  BlocksToCopy = LONG_MAX;		// say essentially unlimited copy
  NoRewind = 0;				// say we're rewinding
  AppendLog = 0;			// and not appending comments
  TwoMarks = 0;				// default to read to EOI

  while( argc > 1)
  {
    if (*argv[0] != '-')
      break;				// if not an option switch

    if ( !strncmp( argv[0], "-n", 2))
    {  // if a block specification
      argc--;
      argv++;				// skip over the switch
      BlocksToCopy = (DWORD) atol( argv[0]);
      if ( !BlocksToCopy)
      {
	fprintf( stderr, "\nError in count specification \"%s\".\n", argv[0]);
	PrintUsage();
      }
      argc--;
      argv++;				// skip over the number
    } // if copy blocks
    else  if ( !strncmp( argv[0], "-r", 2))
    { // if a no-rewind option
      argc--;
      argv++;
      NoRewind = 1;			// say no rewinding
    } // if no rewind
    else  if ( !strncmp( argv[0], "-c", 2))
    { // if a append comments
      argc--;
      argv++;
      AppendLog = 1;		       // say add comments
    } // if no rewind
    else  if ( !strncmp( argv[0], "-t", 2))
    { // if a append comments
      argc--;
      argv++;
      TwoMarks = 1;		      // say add comments
    } // if stop at 2 filemarks
    else
    { // unknown switch
      PrintUsage();
    } // bad option
  } // digest options
 
//  If there's not at least one argument here, print a message and exit.
 
  if ( argc != 1)
    PrintUsage();

//  Get tapes and select one.

  printf( LogonText, ProgramName);
  printf( "\nScanning for tape drives...\n");

  TapeCount = GetTapes();		// inventory the drives

  if ( TapeCount < 0)
  {
    printf( "\nAn ASPI-accessible SCSI controller was not found.\n\n");
    exit(1);
  }

  if ( TapeCount == 0)
  {
    printf( "\nNo ASPI-accessible tape drives were found.\n\n");
    exit(2);
  }

  printf( "\n%d tape drives were found.\n", TapeCount);
  if ( TapeCount == 1)
    SourceTape = 0;			// if only 1, just assume it
  else
  { // we need to select one
    while (TRUE)
    {
      printf( "\nPlease enter the number of the source drive, or 0 to quit -");
      scanf( "%u", addr(SourceTape));
      if ( !SourceTape)
	exit(0);
      SourceTape--;
      if ( SourceTape < TapeCount)
	break;
    } // Get source tape number
  } // if selecting

//  Wait for the tape to become ready

  while (TapeTestReady( 0, SourceTape))
    printf( "Waiting for source tape ready...\n");

//  Allocate a buffer.

  SourceBuffer = _fmalloc( TAPE_BUFFER_SIZE);
  if ( !SourceBuffer)
  {
    printf( "\nNo memory for source tape buffer.\n");
    exit(1);
  }

//  Create the TAP file.  Force the extension to .TAP.

  strcpy(TapFile, argv[0]);
  ch = strrchr(TapFile, '.');
  if ( ch)
    *ch = 0;				// knock the extension off
  strcat(TapFile, TAP_FILE_TYPE);	// Make the image file

  Tap = fopen( TapFile, "wb");
  if (!Tap)
  {
    fprintf( stderr, "\nError - could not create %s.\n\n", TapFile);
    exit(-2);
  } // if unable to open

//  Get the Log	 file and write some stuff to it.

  strcpy(LogFile, argv[0]);
  ch = strrchr(LogFile, '.');
  if ( ch)
    *ch = 0;				// knock the extension off
  strcat(LogFile, LOG_FILE_TYPE);	// Make the log file

  Log = fopen( LogFile, "w");;		// create it
  if (!Log)
  {
    fprintf(stderr, "Could not create log file.\n");
    exit(-8);
  } // unlikley

//  Write some stuff to the temporary file.

  time( &currTime);
  fprintf( Log,
    "\n%s created by %s %s on %s\n"
    "Options are as follows:\n",
     TapFile, ProgramName, VERSION, ctime( &currTime));

//  Show options.  First, how about rewinding?

  fprintf( Log, "  %s at beginning and end.\n",
    NoRewind ? "No rewind" : "Rewind");

//  Next, see if there's a limit on blocks.

  if ( BlocksToCopy != LONG_MAX)
    fprintf( Log, "  Copy %ld blocks.\n", BlocksToCopy);
  else
    fprintf( Log, "  Copy all blocks.\n");

//  How about stopping at 2 tapemarks?

  fprintf( Log, "  Copy until %s.\n",
    TwoMarks ? "double tapemark" : "EOD");

//  Finally, do we append the log?

  if (AppendLog)
    fprintf( Log, "  Appending log  after EOT.\n");

//  Get the comment from the console and write it to the log file.

  printf(
    "\nEnter comment for this tape; enter last tine as a single \".\"\n");

  Logprintf( "%s", COMMENT_SEPARATOR);
  while (TRUE)
  {
    fgets( commentLine, COMMENT_LINE_MAX, stdin);
    if ( !strcmp( commentLine, ".\n"))
      break;				// if a lone ".", quit
    fputs( commentLine, Log);
  }
  Logprintf( "%s", COMMENT_SEPARATOR);

  if ( !NoRewind)
  {
    Logprintf( "\nRewinding tape.\n");
    TapeRewind( SourceTape);		  // rewind the tape
    while (TapeTestReady( 0, SourceTape));
  } // if we need to rewind
  Logprintf( "\nReading...\n\n");

  tapemarkHit = 0;			// say no tapemark to start
  BytesCopied = 0L;			// clear the count
  LastRecordLength = 0xffff;		// clear record lengtth and count
  LastRecordCount = 0;

//  Read the tape and write to disk.

  while (BlocksToCopy)
  { // read a bunch
    status = TapeReadVariable( SourceBuffer, TAPE_BUFFER_SIZE,
			       addr(ActualRead), SourceTape);

    if ( status == SCSI_SENSE_NO_SENSE)
    { // if we have a good read

//  Do some scanning stuff.

      if ( LastRecordLength == ActualRead)
	LastRecordCount++;
      else
      {
	FlushRecordCount();
	LastRecordLength = ActualRead;		// start a new string
	LastRecordCount = 1;	      // say we have one
      } // starting a new string

      tapemarkHit = 0;
      CurrentLength = ActualRead;
      fwrite( &CurrentLength, sizeof( CurrentLength), 1, Tap);	// first count
      fwrite( SourceBuffer, ActualRead, 1, Tap);  // write the tape
      fwrite( &CurrentLength, sizeof( CurrentLength), 1, Tap);	// end count
      BytesCopied += ((DWORD) ActualRead);
    } // if no error
    else
    {  // We get a sense key returned
      switch( status)
      {
	case SCSI_SENSE_NOT_READY:
	  FlushRecordCount();
	  Logprintf( "Drive went not-ready.  Exiting.");
	  break;

	case SCSI_SENSE_FILEMARK:
	  FlushRecordCount();		// note that TAPEIO handles the case
					// of a 1-byte record = tapemark
	  tapemarkHit++;
	  CurrentLength = TAP_FILEMARK;
	  fwrite( &CurrentLength, sizeof( CurrentLength), 1, Tap);  // mark
	  Logprintf( "Tapemark hit at block %lu\n"
		, TapeReadPosition( SourceTape));
	  if ( (tapemarkHit == 2) && TwoMarks)
	    break;
	  status = 0;			 // note the tapemark
	  break;

	case SCSI_SENSE_NO_SENSE:
	  status = 0;			// no error
	  break;
 
	case SCSI_SENSE_MEDIUM_ERROR:
	  FlushRecordCount();
	  Logprintf( "\nMedium error hit at block %lu\n",
		  TapeReadPosition( SourceTape));
	  if (!ActualRead)
	    ActualRead = LastRecordLength ? LastRecordLength : 4;

	  CurrentLength = ActualRead | TAP_ERROR_FLAG;

//  Write count bytes with the ERROR flag set.	If we don't know the bloack
//  length, we use 4.

	  fwrite( &CurrentLength, sizeof( CurrentLength), 1, Tap);
	  fwrite( SourceBuffer, ActualRead, 1, Tap);  // write the tape
	  fwrite( &CurrentLength, sizeof( CurrentLength), 1, Tap);  // end count
	  BytesCopied += ((DWORD) ActualRead);
	  status = 0;
	  break;

	case SCSI_SENSE_BLANK_CHECK:
	  FlushRecordCount();
	  Logprintf( "\nEnd of data hit at %lu\n",
	    TapeReadPosition( SourceTape));
	  break;

	default:
	  FlushRecordCount();
	  Logprintf( "\nError %d hit at block %lu\n", status,
		  TapeReadPosition( SourceTape));
	  break;
      } // switch

//  If we hit a fatal error, we just quit.

      if ( status)
	break;
    } // if we have an exception
    BlocksToCopy--;
  } // while

  FlushRecordCount();
  if ( status == SCSI_SENSE_FILEMARK)
    Logprintf("\nDouble tapemark -- stopping.\n");

  if ( !BlocksToCopy)
    Logprintf("\nSpecified copy count met -- stopping.\n");
 
  CurrentLength = TAP_EOM;    // say EOM
  fwrite( &CurrentLength, sizeof( CurrentLength), 1, Tap);  // mark EOM

  time( &currTime);
  Logprintf("\nFinished reading at %s\n", ctime( &currTime));

//  Okay, now rewind and read the temporary file; append to the image file,
//  if requested.

  if ( AppendLog)
  { // if we need to append the log
    rewind( Log);
    fseek( Log, 0L, SEEK_END);
    CurrentLength = ftell( Log);	 // get length of file

    rewind ( Log);
    fwrite( &CurrentLength, sizeof( CurrentLength), 1, Tap);  // start of comment
    BytesCopied = CurrentLength;	  // get bytes to copy
    while( BytesCopied--)
    {
      putc( getc( Log), Tap);
    }
    fwrite( &CurrentLength, sizeof( CurrentLength), 1, Tap);  // end of comment
  } // if appending comments

//  Finish up

  if ( !NoRewind && (status != SCSI_SENSE_NOT_READY))
  {
    printf( "\nRewinding...\n");
    TapeRewind( SourceTape);
  } // if we rewind at end

  fclose( Tap);
  fclose( Log);
  printf( "\nAll Done - %s and %s written!\n", TapFile, LogFile);
 
  _ffree( SourceBuffer);
  exit(0);
}  // end of main


//  FlushRecordCount - Print possible summary of records encountered.
//  -----------------------------------------------------------------
//
//    Summarizes and resets the counter.
//

void FlushRecordCount( void)
{

  if ( LastRecordCount)
  { // got some; report on it

    Logprintf( "%u x %u bytes\n", LastRecordCount, LastRecordLength);
    LastRecordCount = 0;
    LastRecordLength = 0xffff;		// clear the counters
  }
  return;
} // FlushRecordCount

//  Logprintf - Print to both stdout and the temp file.
//  ---------------------------------------------------
//
//  Saves us the trouble of reporting both.
//

void Logprintf( char *Format, ...)
{
  va_list argl;

  va_start( argl, Format);
  vprintf( Format, argl);
  vfprintf( Log, Format, argl);
  va_end( argl);
  return;
} // Logprintf

//  PrintUsage - Prints Usage/Syntax Information and Exits.
//  ------------------------------------------------------
//
//    No arguments--does not return.
//

void PrintUsage( void)
{

  fprintf( stderr,UsageText, ProgramName, ProgramName);
  exit( -1);
} // PrintUsage
